#include <time.h>
#include <math.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdbool.h>
#include <windows.h>

#include "65PR02.h"


#define MAX_TICKS		100
uint32_t tickIndex = 0;
double tickSum = 0;
double tickList[MAX_TICKS];
double PCFreq;
uint64_t CounterStart;

double AverageTick(double newtick){
	tickSum -= tickList[tickIndex];
	tickSum += newtick;
	tickList[tickIndex] = newtick;
	if(++tickIndex == MAX_TICKS){
		tickIndex = 0;
	}
	return(tickSum / MAX_TICKS);
}


void startCounter(){
	LARGE_INTEGER li;
	QueryPerformanceFrequency(&li);
	PCFreq = ((double)li.QuadPart) / 1000.0f;
	QueryPerformanceCounter(&li);
	CounterStart = li.QuadPart;
}

double getCounter(){
	LARGE_INTEGER li;
	QueryPerformanceCounter(&li);
	return ((double)li.QuadPart-CounterStart) / PCFreq;
}



uint8_t buftemp = 0;
uint64_t bufload = 0;

int main(int argc, char* argv[]){
	time_t t;
	srand((unsigned) time(&t));
	
	cpu_state CPU;
	
	uint8_t *ROM;
	uint32_t ROMsize = readROM(&ROM, "EMU.bin");
	
	
	init(&CPU, ROM, ROMsize);	// Always takes 5 Cycles, so start counting from there
	
	uint64_t bufacc = 0, cycles = 5, instructions = 0, ctrlflow = 0;
	uint16_t rCycles, bufmin = BUF_SIZE, bufmax = 0;
	uint16_t prevPC = CPU.reg_pc;
	
	startCounter();
	while(1){
		rCycles = execute(&CPU, false);
		if (rCycles == 127) goto exit;
		if (prevPC == CPU.reg_pc){
			printf("\nCPU Stuck in a loop at $%04X!\n", CPU.reg_pc);
			exit(EXIT_FAILURE);
		} else {
			prevPC = CPU.reg_pc;
		}
		if (rCycles == 0){
			printf("\nInvalid Instruction!\n");
			exit(EXIT_FAILURE);
		}
		
		if (rCycles & 0x80){	// Check for the Highest bit, if it's set, a Jump/Branch/etc was executed
			rCycles &= 0x7F;
			ctrlflow++;
		}
		
		cycles += rCycles;
		
		
		if (buftemp < bufmin) bufmin = buftemp;
		if (buftemp > bufmax) bufmax = buftemp;
		
		bufacc += buftemp;
		instructions++;
	}
	exit:
	
	printf("\n\nProgram Stopped at PC=$%04X\n", CPU.reg_pc);
	printf("Execution took %.3f Seconds\n", getCounter() / 1000.0f);
	printf("Stats:\n    Executed Instructions: %llu\n    Total Cycles: %llu\n    Average Cycles per Instruction: %.3f\n", instructions, cycles, ((float)cycles / (float)instructions));
	printf("    Amount of TAKEN Control Flow Instructions: %llu (%.3f%% out of all Instructions)\n", ctrlflow, ((float)ctrlflow / (float)instructions) * 100.0f);
	printf("    Lowest Buffer Size: %u\n    Highest Buffer Size: %u\n    Average Buffer Size: %.3f\n", bufmin, bufmax, ((float)bufacc / (float)instructions));
	printf("    Amount of times the Buffer had to be refilled: %llu\n    (%.3f%% of them were from Control Flow Instructions)\n", bufload, ((float)ctrlflow / (float)bufload) * 100.0f);
	
	
	
	free(ROM);
	return 0;
}








// Prints out all Registers
void printstate(cpu_state* CPU){
	printf(state_str, A, X, Y, SP, PC, BTOC(N_FL), BTOC(V_FL), BTOC(D_FL), BTOC(I_FL), BTOC(Z_FL), BTOC(C_FL));
}

// Opens and Reads a file as binary, and stores it's contents into a malloc'd Array
uint32_t readROM(uint8_t** ROM, const char* str){
	FILE *fp;
	size_t fileSize;
	
	fp = fopen(str, "rb");
	if (!fp){
		printf("ROM file not found!\n");
		return 0;
	}
	
	fseek(fp, 0, SEEK_END);
	fileSize = ftell(fp);
	rewind(fp);
	
	if (fileSize > 65536){
		printf("File too large!\n");
		exit(EXIT_FAILURE);
	}
	
	*ROM = malloc(sizeof(uint8_t*) * fileSize);
	if (*ROM == NULL){
		printf("Malloc Failed!\n");
		exit(EXIT_FAILURE);
	}
	
	fread(*ROM, sizeof(uint8_t), fileSize, fp);
	fclose(fp);
	
	return fileSize;
}

// Reset/Initialize the CPU
void init(cpu_state* CPU, uint8_t* ROM, uint32_t ROMsize){
	
	printf("Program is: %u Bytes Large\n", ROMsize);
	// Load the ROM into Memory
	uint32_t ROMstart = 65536 - ROMsize;
	for (uint32_t i = 0; i < ROMsize; i++){
		//printf("%04X: %02X\n", ROMstart + i, ROM[i]);
		switch((ROMstart + i) >> 8){
			case 0:		// Zero Page
				CPU->zp[(ROMstart + i) & 255U] = ROM[i];
			break;
			case 1:		// Stack
				CPU->stack[(ROMstart + i) & 255U] = ROM[i];
			break;
			default:	// Main Memory
				CPU->mem[ROMstart + i] = ROM[i];
			break;
		}
		
	}
	
	// Initialize the Main Register
	A = 0;
	X = 0;
	Y = 0;
	SP = 0;
	SR = SR_INIT;
	
	// Initialize the Program Counter, and Fetch the Starting Address (2 Cycles)
	PC = readmem16(CPU, 0xFFFC, false, false);
	FPC = PC;
	
	printf("CPU Starts at: %04X\n", PC);
	
	// Reset the Instruction Buffer
	resetbuf(CPU);
	
	// And then Load 3 Bytes into it (3 Cycles)
	while (buflen(CPU) < 3){
		writebuf(CPU);
	}
	
}

// Execute a single Instruction, Returns how many cycles it took
uint8_t execute(cpu_state* CPU, bool print){
	// Use the PC to get an Instruction from the Buffer
	uint8_t opc = readbuf(CPU);
	uint8_t tmp0, tmp1, tmp2;
	uint16_t tmp16;
	uint8_t cycles = inst_cycles[opc];
	bool loadPC = false;
	
	if (print) printf("Instruction: %s\t($%04X)\n", inst_names[opc], PC);
	
	// Decode the Addressing Mode and Instruction Type
	switch (opc){
		// Misc
		case NOP:
			FETCH(MEM_NONE);
		break;
		case WAI:
			FETCH(MEM_NONE);	// Non-functional because Interrupts aren't a thing in this Emulator
		break;
		case STP:
			return 127;
		break;
		
		// Jumps and such
		case BRK:
			push16(CPU, (PC+2));
			push8(CPU, (SR | SR_BRK));
			writeflag(CPU, I_FL, true);
			writeflag(CPU, D_FL, false);
			PC = readmem16(CPU, 0xFFFE, false, false);
			loadPC = true;
		break;
		case RTI:
			SR = pull8(CPU) | SR_BRK;
			PC = pull16(CPU);
			loadPC = true;
		break;
		case JSR_A:
			push16(CPU, (PC+2));
			tmp0 = readbuf(CPU);	// Low Byte
			tmp1 = readbuf(CPU);	// High Byte
			PC = COMBINE(tmp1, tmp0);
			loadPC = true;
		break;
		case RTS:
			PC = pull16(CPU) + 1;
			loadPC = true;
		break;
		case JMP_A:
			tmp0 = readbuf(CPU);	// Low Byte
			tmp1 = readbuf(CPU);	// High Byte
			PC = COMBINE(tmp1, tmp0);
			loadPC = true;
		break;
		case JMP_AI:
			tmp0 = readbuf(CPU);	// Low Byte
			tmp1 = readbuf(CPU);	// High Byte
			tmp16 = COMBINE(tmp1, tmp0);
			if ((memtype(tmp16) == MEM_MAIN) && (MEM_SIZE == 1)) cycles++;
			PC = readmem16(CPU, tmp16, false, false);
			loadPC = true;
		break;
		case JMP_AXI:
			tmp0 = readbuf(CPU);	// Low Byte
			tmp1 = readbuf(CPU);	// High Byte
			tmp16 = COMBINE(tmp1, tmp0);
			if ((memtype(tmp16) == MEM_MAIN) && (MEM_SIZE == 1)) cycles++;
			PC = readmem16(CPU, tmp16 + X, false, false);
			loadPC = true;
		break;
		case BRA_R:
			tmp0 = readbuf(CPU);
			PC += ((int16_t)(int8_t)tmp0) + 2;
			loadPC = true;
		break;
		
		// Conditional Branches
		case BPL_R:
		case BMI_R:
			tmp0 = readbuf(CPU);	// Relative Offset
			if (!readflag(CPU, N_FL) == !(opc & 0b00100000)){
				PC += ((int16_t)(int8_t)tmp0) + 2;
				loadPC = true;
			}
		break;
		case BVC_R:
		case BVS_R:
			tmp0 = readbuf(CPU);	// Relative Offset
			if (!readflag(CPU, V_FL) == !(opc & 0b00100000)){
				PC += ((int16_t)(int8_t)tmp0) + 2;
				loadPC = true;
			}
		break;
		case BCC_R:
		case BCS_R:
			tmp0 = readbuf(CPU);	// Relative Offset
			if (!readflag(CPU, C_FL) == !(opc & 0b00100000)){
				PC += ((int16_t)(int8_t)tmp0) + 2;
				loadPC = true;
			}
		break;
		case BNE_R:
		case BEQ_R:
			tmp0 = readbuf(CPU);	// Relative Offset
			if (!readflag(CPU, Z_FL) == !(opc & 0b00100000)){
				PC += ((int16_t)(int8_t)tmp0) + 2;
				loadPC = true;
			}
		break;
		case BBR0_ZPR:
		case BBR1_ZPR:
		case BBR2_ZPR:
		case BBR3_ZPR:
		case BBR4_ZPR:
		case BBR5_ZPR:
		case BBR6_ZPR:
		case BBR7_ZPR:
			tmp0 = readbuf(CPU);	// ZP Address
			tmp1 = readbuf(CPU);	// Relative Offset
			if (!(readmem8(CPU, ((uint16_t)tmp0) & 255U, 0) & (1U << ((opc >> 4) & 7U)))){
				PC += ((int16_t)(int8_t)tmp1) + 3;
				loadPC = true;
			}
		break;
		case BBS0_ZPR:
		case BBS1_ZPR:
		case BBS2_ZPR:
		case BBS3_ZPR:
		case BBS4_ZPR:
		case BBS5_ZPR:
		case BBS6_ZPR:
		case BBS7_ZPR:
			tmp0 = readbuf(CPU);	// ZP Address
			tmp1 = readbuf(CPU);	// Relative Offset
			if (readmem8(CPU, ((uint16_t)tmp0) & 255U, 0) & (1U << ((opc >> 4) & 7U))){
				PC += ((int16_t)(int8_t)tmp1) + 3;
				loadPC = true;
			}
		break;
		
		// Stack Push/Pull
		case PHA:
			FETCH(MEM_STACK);
			push8(CPU, A);
		break;
		case PHX:
			FETCH(MEM_STACK);
			push8(CPU, X);
		break;
		case PHY:
			FETCH(MEM_STACK);
			push8(CPU, Y);
		break;
		case PHP:
			FETCH(MEM_STACK);
			push8(CPU, SR | SR_BRK);
		break;
		case PLA:
			FETCH(MEM_STACK);
			A = pull8(CPU);
			SETNZ(A);
		break;
		case PLX:
			FETCH(MEM_STACK);
			X = pull8(CPU);
			SETNZ(X);
		break;
		case PLY:
			FETCH(MEM_STACK);
			Y = pull8(CPU);
			SETNZ(Y);
		break;
		case PLP:
			FETCH(MEM_STACK);
			SR = pull8(CPU) | SR_BRK;
		break;
		
		// Transfers
		case TAX:
			FETCH(MEM_NONE);
			X = A;
			SETNZ(X);
		break;
		case TAY:
			FETCH(MEM_NONE);
			Y = A;
			SETNZ(Y);
		break;
		case TSX:
			FETCH(MEM_NONE);
			X = SP;
			SETNZ(X);
		break;
		case TXA:
			FETCH(MEM_NONE);
			A = X;
			SETNZ(A);
		break;
		case TYA:
			FETCH(MEM_NONE);
			A = Y;
			SETNZ(A);
		break;
		case TXS:
			FETCH(MEM_NONE);
			SP = X;
		break;
		
		// Flag Set/Clear
		case CLC:
			FETCH(MEM_NONE);
			writeflag(CPU, C_FL, false);
		break;
		case CLD:
			FETCH(MEM_NONE);
			writeflag(CPU, D_FL, false);
		break;
		case CLI:
			FETCH(MEM_NONE);
			writeflag(CPU, I_FL, false);
		break;
		case CLV:
			FETCH(MEM_NONE);
			writeflag(CPU, V_FL, false);
		break;
		case SEC:
			FETCH(MEM_NONE);
			writeflag(CPU, C_FL, true);
		break;
		case SED:
			FETCH(MEM_NONE);
			writeflag(CPU, D_FL, true);
		break;
		case SEI:
			FETCH(MEM_NONE);
			writeflag(CPU, I_FL, true);
		break;
		
		// Load A/X/Y
		// Immediate
		case LDA_I:
			FETCH(MEM_NONE);
			A = readbuf(CPU);
			SETNZ(A);
		break;
		case LDX_I:
			FETCH(MEM_NONE);
			X = readbuf(CPU);
			SETNZ(X);
		break;
		case LDY_I:
			FETCH(MEM_NONE);
			Y = readbuf(CPU);
			SETNZ(Y);
		break;
		// Zero Page
		case LDA_ZP:
			tmp0 = readbuf(CPU);
			A = readmem8(CPU, ((uint16_t)tmp0) & 255U, 1);
			SETNZ(A);
		break;
		case LDX_ZP:
			tmp0 = readbuf(CPU);
			X = readmem8(CPU, ((uint16_t)tmp0) & 255U, 1);
			SETNZ(X);
		break;
		case LDY_ZP:
			tmp0 = readbuf(CPU);
			Y = readmem8(CPU, ((uint16_t)tmp0) & 255U, 1);
			SETNZ(Y);
		break;
		// Zero Page,X
		case LDA_ZPX:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);
			A = readmem8(CPU, ((uint16_t)tmp0 + X) & 255U, 1);
			SETNZ(A);
		break;
		case LDY_ZPX:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);
			Y = readmem8(CPU, ((uint16_t)tmp0 + X) & 255U, 1);
			SETNZ(Y);
		break;
		// Zero Page,Y
		case LDX_ZPY:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);
			X = readmem8(CPU, ((uint16_t)tmp0 + Y) & 255U, 1);
			SETNZ(X);
		break;
		// Absolute
		case LDA_A:
			tmp0 = readbuf(CPU);	// Low Byte
			tmp1 = readbuf(CPU);	// High Byte
			tmp16 = COMBINE(tmp1, tmp0);
			A = readmem8(CPU, tmp16, 1);
			SETNZ(A);
		break;
		case LDX_A:
			tmp0 = readbuf(CPU);	// Low Byte
			tmp1 = readbuf(CPU);	// High Byte
			tmp16 = COMBINE(tmp1, tmp0);
			X = readmem8(CPU, tmp16, 1);
			SETNZ(X);
		break;
		case LDY_A:
			tmp0 = readbuf(CPU);	// Low Byte
			tmp1 = readbuf(CPU);	// High Byte
			tmp16 = COMBINE(tmp1, tmp0);
			Y = readmem8(CPU, tmp16, 1);
			SETNZ(Y);
		break;
		// Absolute,X
		case LDA_AX:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// Low Byte
			tmp1 = readbuf(CPU);	// High Byte
			tmp16 = COMBINE(tmp1, tmp0);
			A = readmem8(CPU, tmp16 + X, 1);
			SETNZ(A);
		break;
		case LDY_AX:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// Low Byte
			tmp1 = readbuf(CPU);	// High Byte
			tmp16 = COMBINE(tmp1, tmp0);
			Y = readmem8(CPU, tmp16 + X, 1);
			SETNZ(Y);
		break;
		// Absolute,Y
		case LDA_AY:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// Low Byte
			tmp1 = readbuf(CPU);	// High Byte
			tmp16 = COMBINE(tmp1, tmp0);
			A = readmem8(CPU, tmp16 + Y, 1);
			SETNZ(A);
		break;
		case LDX_AY:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// Low Byte
			tmp1 = readbuf(CPU);	// High Byte
			tmp16 = COMBINE(tmp1, tmp0);
			X = readmem8(CPU, tmp16 + Y, 1);
			SETNZ(X);
		break;
		// Zero Page,X Indirect
		case LDA_ZPXI:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// ZP Address
			tmp16 = readmem16(CPU, ((uint16_t)tmp0 + X) & 255U, true, true);	// Indirect Address
			if (memtype(tmp16) == MEM_MAIN) cycles++;
			A = readmem8(CPU, tmp16, 1);
			SETNZ(A);
		break;
		// Zero Page Indirect,Y
		case LDA_ZPIY:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// ZP Address
			tmp16 = readmem16(CPU, ((uint16_t)tmp0) & 255U, true, true);		// Indirect Address
			if (memtype(tmp16) == MEM_MAIN) cycles++;
			A = readmem8(CPU, tmp16 + Y, 1);
			SETNZ(A);
		break;
		// Zero Page Indirect
		case LDA_ZPI:
			tmp0 = readbuf(CPU);	// ZP Address
			tmp16 = readmem16(CPU, ((uint16_t)tmp0) & 255U, true, true);		// Indirect Address
			if (memtype(tmp16) == MEM_MAIN) cycles++;
			A = readmem8(CPU, tmp16, 1);
			SETNZ(A);
		break;
		
		// Store A/X/Y/Z
		// Zero Page
		case STA_ZP:
			tmp0 = readbuf(CPU);
			writemem8(CPU, A, (uint16_t)tmp0 & 255U, 1);
		break;
		case STX_ZP:
			tmp0 = readbuf(CPU);
			writemem8(CPU, X, (uint16_t)tmp0 & 255U, 1);
		break;
		case STY_ZP:
			tmp0 = readbuf(CPU);
			writemem8(CPU, Y, (uint16_t)tmp0 & 255U, 1);
		break;
		case STZ_ZP:
			tmp0 = readbuf(CPU);
			writemem8(CPU, 0x00, (uint16_t)tmp0 & 255U, 1);
		break;
		// Zero Page,X
		case STA_ZPX:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);
			writemem8(CPU, A, ((uint16_t)tmp0 + X) & 255U, 1);
		break;
		case STY_ZPX:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);
			writemem8(CPU, Y, ((uint16_t)tmp0 + X) & 255U, 1);
		break;
		case STZ_ZPX:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);
			writemem8(CPU, 0x00, ((uint16_t)tmp0 + X) & 255U, 1);
		break;
		// Zero Page,Y
		case STX_ZPY:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);
			writemem8(CPU, X, ((uint16_t)tmp0 + Y) & 255U, 1);
		break;
		// Absolute
		case STA_A:
			tmp0 = readbuf(CPU);	// Low Byte
			tmp1 = readbuf(CPU);	// High Byte
			tmp16 = COMBINE(tmp1, tmp0);
			writemem8(CPU, A, tmp16, 1);
		break;
		case STX_A:
			tmp0 = readbuf(CPU);	// Low Byte
			tmp1 = readbuf(CPU);	// High Byte
			tmp16 = COMBINE(tmp1, tmp0);
			writemem8(CPU, X, tmp16, 1);
		break;
		case STY_A:
			tmp0 = readbuf(CPU);	// Low Byte
			tmp1 = readbuf(CPU);	// High Byte
			tmp16 = COMBINE(tmp1, tmp0);
			writemem8(CPU, Y, tmp16, 1);
		break;
		case STZ_A:
			tmp0 = readbuf(CPU);	// Low Byte
			tmp1 = readbuf(CPU);	// High Byte
			tmp16 = COMBINE(tmp1, tmp0);
			writemem8(CPU, 0x00, tmp16, 1);
		break;
		// Absolute,X
		case STA_AX:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// Low Byte
			tmp1 = readbuf(CPU);	// High Byte
			tmp16 = COMBINE(tmp1, tmp0);
			writemem8(CPU, A, tmp16 + X, 1);
		break;
		case STZ_AX:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// Low Byte
			tmp1 = readbuf(CPU);	// High Byte
			tmp16 = COMBINE(tmp1, tmp0);
			writemem8(CPU, 0x00, tmp16 + X, 1);
		break;
		// Absolute,Y
		case STA_AY:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// Low Byte
			tmp1 = readbuf(CPU);	// High Byte
			tmp16 = COMBINE(tmp1, tmp0);
			writemem8(CPU, A, tmp16 + Y, 1);
		break;
		// Zero Page,X Indirect
		case STA_ZPXI:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// ZP Address
			tmp16 = readmem16(CPU, ((uint16_t)tmp0 + X) & 255U, true, true);	// Indirect Address
			if (memtype(tmp16) == MEM_MAIN) cycles++;
			writemem8(CPU, A, tmp16, 1);
		break;
		// Zero Page Indirect,Y
		case STA_ZPIY:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// ZP Address
			tmp16 = readmem16(CPU, ((uint16_t)tmp0) & 255U, true, true);		// Indirect Address
			if (memtype(tmp16) == MEM_MAIN) cycles++;
			writemem8(CPU, A, tmp16 + Y, 1);
		break;
		// Zero Page Indirect
		case STA_ZPI:
			tmp0 = readbuf(CPU);	// ZP Address
			tmp16 = readmem16(CPU, ((uint16_t)tmp0) & 255U, true, true);		// Indirect Address
			if (memtype(tmp16) == MEM_MAIN) cycles++;
			writemem8(CPU, A, tmp16, 1);
		break;
		
		// ALU Operations, sorted by Addressing Mode
		// Immediate
		case ORA_I:
		case AND_I:
		case XOR_I:
		case ADC_I:
		case CMP_I:
		case SBC_I:
			g1alu(CPU, readbuf(CPU), aaa(opc), 1);
		break;
		// Zero Page
		case ORA_ZP:
		case AND_ZP:
		case XOR_ZP:
		case ADC_ZP:
		case CMP_ZP:
		case SBC_ZP:
			tmp0 = readbuf(CPU);	// ZP Address
			g1alu(CPU, readmem8(CPU, ((uint16_t)tmp0) & 255U, 1), aaa(opc), 1);
		break;
		// Zero Page,X
		case ORA_ZPX:
		case AND_ZPX:
		case XOR_ZPX:
		case ADC_ZPX:
		case CMP_ZPX:
		case SBC_ZPX:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// ZP Address
			g1alu(CPU, readmem8(CPU, ((uint16_t)tmp0 + X) & 255U, 1), aaa(opc), 1);
		break;
		// Absolute
		case ORA_A:
		case AND_A:
		case XOR_A:
		case ADC_A:
		case CMP_A:
		case SBC_A:
			tmp0 = readbuf(CPU);	// Low Byte
			tmp1 = readbuf(CPU);	// High Byte
			tmp16 = COMBINE(tmp1, tmp0);
			g1alu(CPU, readmem8(CPU, tmp16, 1), aaa(opc), 1);
		break;
		// Absolute,X
		case ORA_AX:
		case AND_AX:
		case XOR_AX:
		case ADC_AX:
		case CMP_AX:
		case SBC_AX:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// Low Byte
			tmp1 = readbuf(CPU);	// High Byte
			tmp16 = COMBINE(tmp1, tmp0);
			g1alu(CPU, readmem8(CPU, tmp16 + X, 1), aaa(opc), 1);
		break;
		// Absolute,Y
		case ORA_AY:
		case AND_AY:
		case XOR_AY:
		case ADC_AY:
		case CMP_AY:
		case SBC_AY:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// Low Byte
			tmp1 = readbuf(CPU);	// High Byte
			tmp16 = COMBINE(tmp1, tmp0);
			g1alu(CPU, readmem8(CPU, tmp16 + Y, 1), aaa(opc), 1);
		break;
		// Zero Page,X Indirect
		case ORA_ZPXI:
		case AND_ZPXI:
		case XOR_ZPXI:
		case ADC_ZPXI:
		case CMP_ZPXI:
		case SBC_ZPXI:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// ZP Address
			tmp16 = readmem16(CPU, ((uint16_t)tmp0 + X) & 255U, true, true);	// Indirect Address
			if (memtype(tmp16) == MEM_MAIN) cycles++;
			g1alu(CPU, readmem8(CPU, tmp16, 1), aaa(opc), 1);
		break;
		// Zero Page Indirect,Y
		case ORA_ZPIY:
		case AND_ZPIY:
		case XOR_ZPIY:
		case ADC_ZPIY:
		case CMP_ZPIY:
		case SBC_ZPIY:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// ZP Address
			tmp16 = readmem16(CPU, ((uint16_t)tmp0) & 255U, true, true);		// Indirect Address
			if (memtype(tmp16) == MEM_MAIN) cycles++;
			g1alu(CPU, readmem8(CPU, tmp16 + Y, 1), aaa(opc), 1);
		break;
		// Zero Page Indirect
		case ORA_ZPI:
		case AND_ZPI:
		case XOR_ZPI:
		case ADC_ZPI:
		case CMP_ZPI:
		case SBC_ZPI:
			tmp0 = readbuf(CPU);	// ZP Address
			tmp16 = readmem16(CPU, ((uint16_t)tmp0) & 255U, true, true);		// Indirect Address
			if (memtype(tmp16) == MEM_MAIN) cycles++;
			g1alu(CPU, readmem8(CPU, tmp16, true), aaa(opc), true);
		break;
		
		// RMW ALU Operations, sorted by Addressing Mode
		// Accumulator
		case ASL:
		case ROL:
		case LSR:
		case ROR:
			A = g2alu(CPU, A, aaa(opc), true);
		break;
		// Special Case for INC/DEC Accumulator, since they don't fit the pattern of the others
		case DEC:
			A--;
			SETNZ(A);
		break;
		case INC:
			A++;
			SETNZ(A);
		break;
		// Zero Page
		case ASL_ZP:
		case ROL_ZP:
		case LSR_ZP:
		case ROR_ZP:
		case DEC_ZP:
		case INC_ZP:
			tmp0 = readbuf(CPU);	// ZP Address
			tmp1 = g2alu(CPU, readmem8(CPU, ((uint16_t)tmp0) & 255U, 1), aaa(opc), 1);
			writemem8(CPU, tmp1, ((uint16_t)tmp0) & 255U, 1);
		break;
		// Zero Page,X
		case ASL_ZPX:
		case ROL_ZPX:
		case LSR_ZPX:
		case ROR_ZPX:
		case DEC_ZPX:
		case INC_ZPX:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// ZP Address
			tmp1 = g2alu(CPU, readmem8(CPU, ((uint16_t)tmp0 + X) & 255U, 1), aaa(opc), 1);
			writemem8(CPU, tmp1, ((uint16_t)tmp0 + X) & 255U, 1);
		break;
		// Absolute
		case ASL_A:
		case ROL_A:
		case LSR_A:
		case ROR_A:
		case DEC_A:
		case INC_A:
			tmp0 = readbuf(CPU);	// Low Byte
			tmp1 = readbuf(CPU);	// High Byte
			tmp16 = COMBINE(tmp1, tmp0);
			tmp1 = g2alu(CPU, readmem8(CPU, tmp16, 1), aaa(opc), 1);
			writemem8(CPU, tmp1, tmp16, 1);
		break;
		// Absolute,X
		case ASL_AX:
		case ROL_AX:
		case LSR_AX:
		case ROR_AX:
		case DEC_AX:
		case INC_AX:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// Low Byte
			tmp1 = readbuf(CPU);	// High Byte
			tmp16 = COMBINE(tmp1, tmp0);
			tmp1 = g2alu(CPU, readmem8(CPU, tmp16 + X, 1), aaa(opc), 1);
			writemem8(CPU, tmp1, tmp16 + X, 1);
		break;
		
		// X/Y based ALU Operations
		case INX:
			FETCH(MEM_NONE);
			X++;
			SETNZ(X);
		break;
		case DEX:
			FETCH(MEM_NONE);
			X--;
			SETNZ(X);
		break;
		case CPX_I:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// Immediate
			SETNZ(X - tmp0);
			writeflag(CPU, C_FL, (X >= tmp0));
		break;
		case CPX_ZP:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// ZP Address
			tmp1 = readmem8(CPU, ((uint16_t)tmp0) & 255U, 1);
			SETNZ(X - tmp1);
			writeflag(CPU, C_FL, (X >= tmp1));
		break;
		case CPX_A:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// Low Byte
			tmp1 = readbuf(CPU);	// High Byte
			tmp16 = COMBINE(tmp1, tmp0);
			tmp1 = readmem8(CPU, tmp16, 1);
			SETNZ(X - tmp1);
			writeflag(CPU, C_FL, (X >= tmp1));
		break;
		case INY:
			FETCH(MEM_NONE);
			Y++;
			SETNZ(Y);
		break;
		case DEY:
			FETCH(MEM_NONE);
			Y--;
			SETNZ(Y);
		break;
		case CPY_I:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// Immediate
			SETNZ(Y - tmp0);
			writeflag(CPU, C_FL, (Y >= tmp0));
		break;
		case CPY_ZP:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// ZP Address
			tmp1 = readmem8(CPU, ((uint16_t)tmp0) & 255U, 1);
			SETNZ(Y - tmp1);
			writeflag(CPU, C_FL, (Y >= tmp1));
		break;
		case CPY_A:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// Low Byte
			tmp1 = readbuf(CPU);	// High Byte
			tmp16 = COMBINE(tmp1, tmp0);
			tmp1 = readmem8(CPU, tmp16, 1);
			SETNZ(Y - tmp1);
			writeflag(CPU, C_FL, (Y >= tmp1));
		break;
		
		// Bit Manipulation Instructions
		// BIT
		// Immediate
		case BIT_I:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// Immediate
			writeflag(CPU, Z_FL, !(tmp0 & A));
		break;
		// Zero Page
		case BIT_ZP:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// ZP Address
			tmp1 = readmem8(CPU, ((uint16_t)tmp0) & 255U, 1);
			writeflag(CPU, Z_FL, !(tmp1 & A));
			writeflag(CPU, V_FL, tmp1 & 0b01000000);
			writeflag(CPU, N_FL, tmp1 & 0b10000000);
		break;
		// Zero Page,X
		case BIT_ZPX:
			FETCH(MEM_NONE);
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// ZP Address
			tmp1 = readmem8(CPU, ((uint16_t)tmp0 + X) & 255U, 1);
			writeflag(CPU, Z_FL, !(tmp1 & A));
			writeflag(CPU, V_FL, tmp1 & 0b01000000);
			writeflag(CPU, N_FL, tmp1 & 0b10000000);
		break;
		// Absolute
		case BIT_A:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// Low Byte
			tmp1 = readbuf(CPU);	// High Byte
			tmp16 = COMBINE(tmp1, tmp0);
			tmp1 = readmem8(CPU, tmp16, 1);
			writeflag(CPU, Z_FL, !(tmp1 & A));
			writeflag(CPU, V_FL, tmp1 & 0b01000000);
			writeflag(CPU, N_FL, tmp1 & 0b10000000);
		break;
		// Absolute,X
		case BIT_AX:
			FETCH(MEM_NONE);
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// Low Byte
			tmp1 = readbuf(CPU);	// High Byte
			tmp16 = COMBINE(tmp1, tmp0);
			tmp1 = readmem8(CPU, tmp16 + X, 1);
			writeflag(CPU, Z_FL, !(tmp1 & A));
			writeflag(CPU, V_FL, tmp1 & 0b01000000);
			writeflag(CPU, N_FL, tmp1 & 0b10000000);
		break;
		
		// TRB/TSB:
		// Zero Page
		case TRB_ZP:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// ZP Address
			tmp1 = readmem8(CPU, ((uint16_t)tmp0) & 255U, 1);
			writemem8(CPU, ~A & tmp1, ((uint16_t)tmp0) & 255U, 1);
			writeflag(CPU, Z_FL, !(tmp1 & A));
		break;
		case TSB_ZP:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// ZP Address
			tmp1 = readmem8(CPU, ((uint16_t)tmp0) & 255U, 1);
			writemem8(CPU, A | tmp1, ((uint16_t)tmp0) & 255U, 1);
			writeflag(CPU, Z_FL, !(tmp1 & A));
		break;
		// Absolute
		case TRB_A:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// Low Byte
			tmp1 = readbuf(CPU);	// High Byte
			tmp16 = COMBINE(tmp1, tmp0);
			tmp1 = readmem8(CPU, tmp16, 1);
			writemem8(CPU, ~A & tmp1, tmp16, 1);
			writeflag(CPU, Z_FL, !(tmp1 & A));
		break;
		case TSB_A:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);	// Low Byte
			tmp1 = readbuf(CPU);	// High Byte
			tmp16 = COMBINE(tmp1, tmp0);
			tmp1 = readmem8(CPU, tmp16, 1);
			writemem8(CPU, A | tmp1, tmp16, 1);
			writeflag(CPU, Z_FL, !(tmp1 & A));
		break;
		
		// Set/Clear Bits in Zero Page
		case SMB0_ZP:
		case SMB1_ZP:
		case SMB2_ZP:
		case SMB3_ZP:
		case SMB4_ZP:
		case SMB5_ZP:
		case SMB6_ZP:
		case SMB7_ZP:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);
			tmp1 = readmem8(CPU, ((uint16_t)tmp0) & 255U, 1);
			tmp1 |= 1U << ((opc >> 4) & 7U);
			writemem8(CPU, tmp1, ((uint16_t)tmp0) & 255U, 1);
		break;
		
		case RMB0_ZP:
		case RMB1_ZP:
		case RMB2_ZP:
		case RMB3_ZP:
		case RMB4_ZP:
		case RMB5_ZP:
		case RMB6_ZP:
		case RMB7_ZP:
			FETCH(MEM_NONE);
			tmp0 = readbuf(CPU);
			tmp1 = readmem8(CPU, ((uint16_t)tmp0) & 255U, 1);
			tmp1 &= ~((uint8_t)(1U << ((opc >> 4) & 7U)));
			writemem8(CPU, tmp1, ((uint16_t)tmp0) & 255U, 1);
		break;
		
		
		default:
			printf("!!!!!!!!!!!!\nInvalid Opcode: 0x%02X\nat Address: %04X\n!!!!!!!!!!!!\n", opc, PC);
			return 0;
		break;
	}
	
	// when loadPC is false, advance the PC and read pointer
	if (loadPC){
		resetbuf(CPU);
		FPC = PC;
	} else {
		PC += inst_size[opc];
	}
	
	// Update the Temporary Buffer Counter
	buftemp = buflen(CPU);
	
	// check if there are atleast 3 Bytes in the Buffer, if not, load some!
	if (buflen(CPU) < 3) bufload++;
	while (buflen(CPU) < 3){
		if (writebuf(CPU)) cycles++;
		if (print) printf("Bytes in Buffer: %u\n", buflen(CPU));
	}
	
	return cycles | (loadPC ? 0x80 : 0x00);
}





// Group 1 ALU Operation
// Accumulator = Accumulator <ALU> input
void g1alu(cpu_state* CPU, uint8_t input, uint8_t operation, bool fetch){
	if (fetch) FETCH(MEM_NONE);
	uint8_t tmp8;
	uint8_t tmpin = input;
	uint16_t tmp16;
	
	switch (operation & 7U){
		case OP_SBC:
			if (readflag(CPU, D_FL)){
				tmpin = ncompl[input];
			} else {
				tmpin = ~input;
			}
		case OP_ADC:
			if (readflag(CPU, D_FL)){
				tmp8 = (A & 0x0F) + (tmpin & 0x0F) + (readflag(CPU, C_FL) ? 1 : 0);
				if (tmp8 > 9) tmp8 += 6;
				
				tmp16 = (A & 0xF0) + (tmpin & 0xF0) + (tmp8 & 0xF0);
				if (tmp16 > 0x90) tmp16 += 0x60;
				writeflag(CPU, C_FL, tmp16 >> 8);
				writeflag(CPU, V_FL, ((~(A ^ tmpin)) & (A ^ tmp16)) & 0x80);
				
				A = (tmp16 & 0x00FF) + (tmp8 & 0x0F);
				SETNZ(A);
			} else {
				tmp16 = (uint16_t)A + tmpin + (readflag(CPU, C_FL) ? 1 : 0);
				writeflag(CPU, C_FL, tmp16 >> 8);
				writeflag(CPU, V_FL, ((~(A ^ tmpin)) & (A ^ tmp16)) & 0x80);
				A = (uint8_t)tmp16;
				SETNZ(A);
			}
		break;
		case OP_CMP:
			tmp8 = A - input;
			SETNZ(tmp8);
			writeflag(CPU, C_FL, (A >= input));
		break;
		case OP_AND:
			A &= input;
			SETNZ(A);
		break;
		case OP_ORA:
			A |= input;
			SETNZ(A);
		break;
		case OP_XOR:
			A ^= input;
			SETNZ(A);
		break;
	}
}

// Group 2 ALU Operation
// output = input <ALU>
uint8_t g2alu(cpu_state* CPU, uint8_t input, uint8_t operation, bool fetch){
	if (fetch) FETCH(MEM_NONE);
	
	uint8_t tmp8 = 0;
	
	switch (operation & 7U){
		case OP_ASL:
			tmp8 = input << 1;
			writeflag(CPU, C_FL, input & 0b10000000);
			SETNZ(tmp8);
		break;
		case OP_LSR:
			tmp8 = input >> 1;
			writeflag(CPU, C_FL, input & 1U);
			SETNZ(tmp8);
		break;
		case OP_ROL:
			tmp8 = input << 1;
			tmp8 |= readflag(CPU, C_FL) ? 1U : 0U;
			writeflag(CPU, C_FL, input & 0b10000000);
			SETNZ(tmp8);
		break;
		case OP_ROR:
			tmp8 = input >> 1;
			tmp8 |= readflag(CPU, C_FL) ? 0b10000000 : 0U;
			writeflag(CPU, C_FL, input & 1U);
			SETNZ(tmp8);
		break;
		case OP_INC:
			tmp8 = input + 1;
			SETNZ(tmp8);
		break;
		case OP_DEC:
			tmp8 = input - 1;
			SETNZ(tmp8);
		break;
	}
	return tmp8;
}

// Gets the current amount of Bytes in the Buffer
uint16_t buflen(cpu_state* CPU){
	if (CPU->buf_full) return BUF_SIZE;
	if (CPU->buf_wrp >= CPU->buf_rdp){
		return CPU->buf_wrp - CPU->buf_rdp;
	} else{
		return (BUF_SIZE + CPU->buf_wrp) - CPU->buf_rdp;
	}
}

// Pulls an 8-bit Value from the Stack
uint8_t pull8(cpu_state* CPU){
	return CPU->stack[++SP];
}

// Pulls a 16-bit Value from the Stack
uint16_t pull16(cpu_state* CPU){
	uint8_t tmp0 = CPU->stack[++SP];
	return tmp0 | ((uint16_t)CPU->stack[++SP] << 8);
}

// Pushes an 8-bit Value to the Stack
void push8(cpu_state* CPU, uint8_t value){
	CPU->stack[SP--] = value;
}

// Pushes a 16-bit Value to the Stack
void push16(cpu_state* CPU, uint16_t value){
	CPU->stack[SP--] = (value >> 8) & 255U;
	CPU->stack[SP--] = value & 255U;
}

// Checks which type of Memory an address would access
// 0 = ZP, 1 = Stack, 2 = Main Memory
uint8_t memtype(uint16_t address){
	switch (address >> 8){
		case 0:		// Zero Page
			return MEM_ZP;
		break;
		case 1:		// Stack
			return MEM_STACK;
		break;
		default:	// Main Memory
			return MEM_MAIN;
		break;
	}
}

// Reads an 8-bit Value from Memory
// if fetch is true, the fecthing unit will also try to read a byte into the Instruction Buffer
uint8_t readmem8(cpu_state* CPU, uint16_t address, bool fetch){
	if (fetch && (memtype(address) != memtype(FPC))){				// this checks if fetching is enabled and the FPC doesn't access the same memory as the CPU
		writebuf(CPU);
	}
	
	if (address == CHAR_IN) return 0;//getchar();
	
	uint8_t tmp = 0;
	switch (address >> 8){
		case 0:		// Zero Page
			tmp = CPU->zp[address & 255U];
			//printf("Reading $%02X from Zero Page: $%04X\n", tmp, address);
		break;
		case 1:		// Stack
			//printf("Reading $%02X from the Stack: $%04X\n", tmp, address);
			tmp = CPU->stack[address & 255U];
		break;
		default:	// Main Memory
			tmp = CPU->mem[address];
		break;
	}
	
	
	return tmp;
}

// Reads a 16-bit Value from Memory
uint16_t readmem16(cpu_state* CPU, uint16_t address, bool wrap, bool fetch){
	if (fetch && (memtype(address) != memtype(FPC))){					// this checks if fetching is enabled and the FPC doesn't access the same memory as the CPU
		if ((memtype(address) == MEM_MAIN) && !(address & ((1U << (MEM_SIZE>>1)) - 1))){
			writebuf(CPU);
		}
		writebuf(CPU);
	}
	
	uint8_t tmp0, tmp1;
	uint16_t tmp16 = address + 1;
	switch (address >> 8){
		case 0:		// Zero Page
			tmp0 = CPU->zp[address & 255U];
			if (wrap){
				tmp1 = CPU->zp[(address + 1) & 255U];
				return COMBINE(tmp1, tmp0);
			}
		break;
		case 1:		// Stack
			tmp0 = CPU->stack[address & 255U];
			if (wrap){
				tmp1 = CPU->stack[(address + 1) & 255U];
				return COMBINE(tmp1, tmp0);
			}
		break;
		default:	// Main Memory
			tmp0 = CPU->mem[address];
		break;
	}
	
	switch (tmp16 >> 8){
		case 0:		// Zero Page
			tmp1 = CPU->zp[tmp16 & 255U];
		break;
		case 1:		// Stack
			tmp1 = CPU->stack[tmp16 & 255U];
		break;
		default:	// Main Memory
			tmp1 = CPU->mem[tmp16];
		break;
	}
	return COMBINE(tmp1, tmp0);
}

// Writes an 8-bit Value to Memory
void writemem8(cpu_state* CPU, uint8_t data, uint16_t address, bool fetch){
	if (fetch && (memtype(address) != memtype(FPC))){				// this checks if fetching is enabled and the FPC doesn't access the same memory as the CPU
		writebuf(CPU);
	}
	if (address == 0xF008){
		printf("$%02X\n", data);
	} else if (address == CHAR_OUT){
		printf("%c", data);
	} //else printf("Writing $%02X to Address: $%04X\n", data, address);
	
	
	
	switch (address >> 8){
		case 0:		// Zero Page
			CPU->zp[address & 255U] = data;
		break;
		case 1:		// Stack
			CPU->stack[address & 255U] = data;
		break;
		default:	// Main Memory
			CPU->mem[address] = data;
		break;
	}
}

// Reads a Byte from the Buffer and moves the Read Pointer, returns -1 if it failed
int16_t readbuf(cpu_state* CPU){
	if ((CPU->buf_wrp == CPU->buf_rdp) && !CPU->buf_full) return -1;
	uint8_t tmp = CPU->buf[CPU->buf_rdp];
	CPU->buf_rdp++;
	//if (CPU->buf_rdp == BUF_SIZE) CPU->buf_rdp = 0;
	CPU->buf_full = false;
	return ((int16_t)tmp) & 255U;
}

// Writes a Byte to the Buffer and moves the Write Pointer, returns "false" if it failed
bool writebuf(cpu_state* CPU){
	uint16_t mask = (1U << (MEM_SIZE>>1)) - 1;
	uint16_t bottom = FPC & mask;
	uint8_t count = 0;
	//printf("Start fetching! (bottom: %X)\n", bottom);
	do {
		//printf("Fetch Cycle #%u, PC = $%04X\n", count, FPC);
		if (CPU->buf_full) return false;
		CPU->buf[CPU->buf_wrp] = readmem8(CPU, FPC, false);
		FPC++;
		CPU->buf_wrp++;
		if (CPU->buf_wrp == CPU->buf_rdp) CPU->buf_full = true;
		count++;
	} while (FPC & mask);
	//printf("Stop fetching! (PC=$%04X)\n", FPC);
	return true;
}

// Resets the Buffer to 0 (ie move both Pointers to 0)
void resetbuf(cpu_state* CPU){
	CPU->buf_wrp = 0;
	CPU->buf_rdp = 0;
	CPU->buf_full = false;
}

// 0 + Carry Flag
// 1 + Zero Flag
// 2 + Interrupt Flag
// 3 + Decimal Flag
// 4 - Break "Flag" (Unused)
// 5 - Unused
// 6 + Overflow Flag
// 7 + Negative Flag
// Reads a Flag from the SR
bool readflag(cpu_state* CPU, uint8_t flag){
	return ((SR >> flag) & 1U);
}

// Writes value (0 or 1) to the selected Flag, if Break or the unused flag are selected, nothing is written
void writeflag(cpu_state* CPU, uint8_t flag, bool value){
	if ((flag == B_FL) || (flag == U_FL)) return;
	value ? (SR |= (1U << flag)) : (SR &= ~(1U << flag));
}























